from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class ImageFeaturesBad(Rule):
    def __init__(self) -> None:
        self._map = {
            'allow-empty-password': ('warning', 'that is a security risk. Users can have accounts without password protection'),
            'allow-root-login': ('info', 'that is a security risk. Root can login into the system'),
            'dbg-pkgs': ('info', 'that exposes sensitive information. Use debuginfod instead'),
            'debug-tweaks': ('warning', 'that is a security risk. Allows limitless access without authentication'),
            'dev-pkgs': ('info', 'that is not recommended, as it bloats the image. Do not use in production'),
            'eclipse-debug': ('info', 'installs debugging tools that could expose sensitive information. Do not use in production'),
            'empty-root-password': ('warning', 'that is a security risk. Anyone can authenticate as root over serial without a password'),
            'post-install-logging': ('info', 'that could expose sensitive information. Not recommended for production'),
            'ptest-pkgs': ('info', 'that bloats the image. Never use in production'),
            'serial-autologin-root': ('warning', 'that is a security risk. Automatically logs in a root'),
            'staticdev-pkgs': ('info', 'that is not recommended, as it bloats the image. Do not use in production'),
            'tools-debug': ('info', 'installs debugging tools that could expose sensitive information. Do not use in production'),
        }
        super().__init__(id='oelint.var.badimagefeature',
                         severity='warning',
                         message="{var} contains '{feature}', {reason}",
                         appendix=list(self._map.keys()))

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        for item in stash.GetItemsFor(filename=_file,
                                      classifier=Variable.CLASSIFIER,
                                      attribute=Variable.ATTR_VAR,
                                      attributeValue=("IMAGE_FEATURES", "EXTRA_IMAGE_FEATURES")):
            if 'remove' in item.SubItems:
                continue
            items = item.get_items()
            for k, v in self._map.items():
                if k in items:
                    severity, reason = v
                    res += self.finding(item.Origin, item.InFileLine,
                                        self.Msg.format(var=item.VarName, feature=k, reason=reason),
                                        appendix=k,
                                        severity_override=severity)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash
from oelint_parser.constants import CONSTANTS

from oelint_adv.cls_rule import Rule


class VarsOrder(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.var.order',
                         severity='warning',
                         message='<FOO>',
                         appendix=[self.__cleanname(x) for x in CONSTANTS.GetByPath('variables/order')])

    def __cleanname(self, _input: str) -> str:
        return _input.replace('$', '').replace('{', '').replace('}', '')

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        _files = {item.Origin for item in stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER)}
        for _single_file in _files:
            items: List[Variable] = stash.GetItemsFor(filename=_single_file, classifier=Variable.CLASSIFIER, nolink=True)
            for item in items:
                _func_before = sorted(
                    [x for x in items if x.Line < item.Line and x.VarName in CONSTANTS.GetByPath('variables/order')], key=lambda x: x.Line, reverse=True)
                if any(_func_before):
                    _func_before = _func_before[0]
                    if item.VarName not in CONSTANTS.GetByPath('variables/order'):
                        continue
                    if CONSTANTS.GetByPath('variables/order').index(item.VarName) < CONSTANTS.GetByPath('variables/order').index(_func_before.VarName):
                        res += self.finding(item.Origin, item.InFileLine,
                                            "'{a}' should be placed before '{b}'".format(
                                                a=item.VarName, b=_func_before.VarName), appendix=self.__cleanname(item.VarName))
        return res



import ast
import textwrap
from typing import List, Tuple

from oelint_parser.cls_item import Function
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class TaskNoPythonPrefix(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.task.nopythonprefix',
                         severity='warning',
                         message="Tasks that don't contain valid python code (e.g. shell code), should not be prefixed with python in function header")

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Function] = stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER)
        for item in items:
            try:
                ast.parse(textwrap.dedent(item.FuncBodyRaw.rstrip('}\n')), 'tempfile')
            except Exception:
                if item.IsPython:
                    res += self.finding(item.Origin, item.InFileLine)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash
from oelint_parser.parser import INLINE_BLOCK

from oelint_adv.cls_rule import Rule


class VarSRCUriGitTag(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.srcurigittag',
                         severity='warning',
                         message="'tag' in SRC_URI-options leads to not-reproducible builds as git-tags can move around. Use explicit SRCREV")

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                  attribute=Variable.ATTR_VAR, attributeValue='SRC_URI')
        for item in items:
            lines = [y.strip("'") for y in item.get_items() if y and INLINE_BLOCK not in y]
            for x in lines:
                _url = stash.GetScrComponents(x)
                if _url['scheme'] in ['git'] and 'tag' in _url['options']:
                    res += self.finding(item.Origin, item.InFileLine)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Function, PythonBlock, TaskAdd, TaskDel
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class TaskDash(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.task.dash',
                         severity='error',
                         message="Task '{FUNC}' shall not have '-' in its name")

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        for item in stash.GetItemsFor(filename=_file, classifier=(
                TaskAdd.CLASSIFIER, Function.CLASSIFIER,
                PythonBlock.CLASSIFIER, TaskDel.CLASSIFIER)):
            value = item.FuncName
            if '-' in value:
                res += self.finding(item.Origin, item.InFileLine,
                                    self.Msg.replace('{FUNC}', value))
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Inherit, Variable
from oelint_parser.cls_stash import Stash
from oelint_parser.constants import CONSTANTS

from oelint_adv.cls_rule import Rule, Classification


class VarSuggestedExists(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.var.suggestedvar',
                         severity='info',
                         message='<FOO>',
                         run_on=[Classification.RECIPE],
                         appendix=CONSTANTS.GetByPath('variables/suggested'))

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        _is_pkg_group = stash.IsPackageGroup(_file)
        _is_image = stash.IsImage(_file)
        all_items: List[Variable] = stash.GetItemsFor(filename=_file,
                                                      classifier=Variable.CLASSIFIER,
                                                      attribute=Variable.ATTR_VAR,
                                                      attributeValue=CONSTANTS.GetByPath('variables/suggested'))
        inherits: List[Inherit] = stash.GetItemsFor(filename=_file,
                                                    classifier=Inherit.CLASSIFIER)

        def get_class_specific(varname: str, inherits: List[Inherit]) -> bool:
            return any(True for x in inherits if any(y in x.get_items()
                                                     for y in (CONSTANTS.GetByPath(f'oelint-suggestedvar/{varname}-exclude-classes') or [])))
        for var_ in CONSTANTS.GetByPath('variables/suggested'):
            if var_ == 'BBCLASSEXTEND':
                # this is better covered by oelint_adv/rule_base/rule_vars_bbclassextends.py
                continue
            if _is_pkg_group and var_ in CONSTANTS.GetByPath('oelint-suggestedvar/pkggroup-excludes'):
                continue
            if _is_image and var_ in CONSTANTS.GetByPath('oelint-suggestedvar/image-excludes'):
                continue
            if get_class_specific(var_, inherits):
                continue
            items = [x for x in all_items if x.VarName == var_]
            if not any(items):
                res += self.finding(_file, 0, "Variable '{var}' should be set".format(var=var_), appendix=var_)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Inherit
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class VarInherit(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.var.inherit',
                         severity='warning',
                         message='<FOO>',
                         appendix=['inherit', 'inherit_defer'],
                         valid_from_release='scarthgap')

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Inherit] = stash.GetItemsFor(filename=_file, classifier=Inherit.CLASSIFIER)
        for i in items:
            if '${' not in i.RealRaw and i.Statement == 'inherit_defer':
                if any(x in i.get_items() for x in ['native', 'nativesdk', 'cross']):
                    continue
                res += self.finding(i.Origin, i.InFileLine,
                                    'inherit_defer should only be used if there is a variable involved',
                                    appendix='inherit')
            if '${' in i.RealRaw and i.Statement == 'inherit':
                res += self.finding(i.Origin, i.InFileLine,
                                    'Variable detected in inherit statement. Use inherit_defer instead',
                                    appendix='inherit_defer')
        return res



import argparse
import copy
import fnmatch
import itertools
import json
import multiprocessing as mp
import os
import re
from configparser import ConfigParser, NoOptionError, NoSectionError, ParsingError
from functools import partial
from typing import Dict, Iterable, List, Tuple, Union

from oelint_parser.cls_item import Comment, Item
from oelint_parser.cls_stash import Stash
from oelint_parser.constants import CONSTANTS
from oelint_parser.rpl_regex import RegexRpl

from oelint_adv.caches import Caches, __default_cache_dir
from oelint_adv.cls_rule import Rule, load_rules
from oelint_adv.rule_base.rule_file_inlinesuppress_na import (
    FileNotApplicableInlineSuppression,
)
from oelint_adv.state import State
from oelint_adv.tweaks import Tweaks

config_basepath = os.getcwd()


class TypeSafeAppendAction(argparse.Action):

    def __call__(self, parser, namespace, values, option_string=None) -> None:
        if not isinstance(values, str):
            return  # pragma: no cover
        items = getattr(namespace, self.dest) or []
        if not isinstance(items, list):
            items = [x.strip() for x in items.split() if x.strip()]
        items.extend(RegexRpl.split(r'\s+|\t+|\n+',
                     values.strip('"').strip("'")))
        setattr(namespace, self.dest, items)


def deserialize_boolean_options(options: Dict) -> Dict[str, Union[str, bool]]:
    """Converts strings in `options` that are either 'True' or 'False' to their boolean
    representations.
    """
    for k, v in options.items():
        if isinstance(v, str):
            if v.strip() == 'False':
                options[k] = False
            elif v.strip() == 'True':
                options[k] = True

    return options


def parse_configfile() -> Dict:
    global config_basepath

    if os.environ.get('OELINT_SKIP_CONFIG', ''):
        return {}
    config = ConfigParser()
    for conffile, basepath in [
        (os.environ.get('OELINT_CONFIG', '/does/not/exist'),
         os.path.dirname(os.environ.get('OELINT_CONFIG', '/does/not/exist'))),
        (os.path.join(os.getcwd(), '.oelint.cfg'), os.getcwd()),
        (os.path.join(os.environ.get('HOME', '/does/not/exist'),
         '.oelint.cfg'), os.environ.get('HOME', '/does/not/exist')),
    ]:
        try:
            if not os.path.exists(conffile):
                continue
            config.read(conffile)
            items = {k.replace('-', '_'): v for k, v in config.items('oelint')}
            items = deserialize_boolean_options(items)

            config_basepath = basepath
            return items
        except (PermissionError, SystemError) as e:  # pragma: no cover
            print(f'Failed to load config file {conffile}. {e!r}')  # noqa: T201 - it's fine here; # pragma: no cover
        except (NoSectionError, NoOptionError, ParsingError) as e:
            print(f'Failed parsing config file {conffile}. {e!r}')  # noqa: T201 - it's here for a reason

    return {}


def group_files(files: Iterable[str], mode: str) -> List[Tuple[List[str], List[str], Dict]]:
    # in case multiple bb files are passed at once we might need to group them to
    # avoid having multiple, potentially wrong hits of include files shared across
    # the bb files in the stash
    res = {}
    for f in files:
        _filename, _ext = os.path.splitext(f)
        if _ext not in ['.bb']:
            continue
        if '_' in os.path.basename(_filename):
            _filename_key = _filename
        else:
            _filename_key = os.path.basename(_filename)
        if _filename_key not in res:  # pragma: no cover
            res[_filename_key] = set()
        res[_filename_key].add(f)

    # second round now for the bbappend files
    for f in files:
        _filename, _ext = os.path.splitext(f)
        if _ext not in ['.bbappend']:
            continue
        _match = False
        for _, v in res.items():
            _needle = '^.*/' + \
                re.escape(os.path.basename(_filename)).replace('%', '.*') + '.bb$'
            if any(RegexRpl.match(_needle, x, re.MULTILINE) for x in v):
                v.add(f)
                _match = True
        if not _match:
            _filename_key = os.path.basename(_filename).replace('%', '')
            if _filename_key not in res:  # pragma: no cover
                res[_filename_key] = set()
            res[_filename_key].add(f)

    # third round for lone bbclasses
    for f in files:
        _filename, _ext = os.path.splitext(f)
        if _ext not in ['.bbclass']:
            continue
        if '_' in os.path.basename(_filename):
            _filename_key = _filename
        else:
            _filename_key = os.path.basename(_filename)
        if _filename_key not in res:  # pragma: no cover
            res[_filename_key] = set()
        res[_filename_key].add(f)

    # layer.confs
    _conf_layer = sorted([x for x in files if os.path.basename(
        x) == 'layer.conf'], key=lambda index: files.index(index))

    _product_matrix = []
    _conf_machine = []
    _conf_distro = []

    # as sets are unordered, we convert them to sorted lists at this point
    # order is like the files have been passed via CLI
    for k, v in res.items():
        res[k] = sorted(v, key=lambda index: files.index(index))

    if mode in ['all']:
        # machine.confs
        _conf_machine = [x for x in files if fnmatch.fnmatch(
            os.path.abspath(x), '*/machine/*.conf')]
        if any(_conf_machine):
            _product_matrix.append(_conf_machine)

        # distro.confs
        _conf_distro = [x for x in files if fnmatch.fnmatch(
            os.path.abspath(x), '*/distro/*.conf')]
        if any(_conf_distro):
            _product_matrix.append(_conf_distro)

        # pos and neg branch expansion
        _product_matrix.append([True, False])
    else:
        _product_matrix.append([False])

    group_res = []

    for fg in res.values():
        _fl = sorted(fg, key=lambda index: files.index(index))
        for element in itertools.product(*_product_matrix):
            _branch_expansion = element[-1]
            _matrix_keys = [
                f'branch:{"false" if _branch_expansion else "true"}']
            if len(element) > 1:
                _files = _conf_layer + list(element[:-1]) + _fl

                _machine_id = [x for x in _files if x in _conf_machine]
                _distro_id = [x for x in _files if x in _conf_distro]

                if any(_machine_id):
                    _matrix_keys.append(os.path.basename(_machine_id[0]))

                if any(_distro_id):
                    _matrix_keys.append(os.path.basename(_distro_id[0]))
            else:
                _files = _conf_layer + _fl
            group_res.append((_files, frozenset(_matrix_keys), {
                             'negative_inline': _branch_expansion}))

    if _conf_layer:
        if mode in ['all']:
            group_res.append((_conf_layer, frozenset(
                ['branch=true']), {'negative_inline': True}))
        group_res.append((_conf_layer, frozenset(
            ['branch=false']), {'negative_inline': False}))

    for m in _conf_machine:
        # mode == all is implicit at this stage
        group_res.append(([m], frozenset(['branch=true']),
                         {'negative_inline': True}))
        group_res.append(([m], frozenset(['branch=false']),
                         {'negative_inline': False}))

    for d in _conf_distro:
        # mode == all is implicit at this stage
        group_res.append(([d], frozenset(['branch=true']),
                         {'negative_inline': True}))
        group_res.append(([d], frozenset(['branch=false']),
                         {'negative_inline': False}))

    return group_res


def group_run(group: List[Tuple],
              quiet: bool,
              fix: bool,
              rules: List[Rule],
              state: State,
              persistent: bool = True) -> Tuple[List[Tuple[str, int, str]], Dict[str, str]]:
    fixedfiles = []
    group_files, matrix, stash_params = group
    stash = Stash(
        quiet=quiet, **state.get_additional_stash_args(), **stash_params)

    for f in group_files:
        try:
            stash.AddFile(f)
        except FileNotFoundError as e:  # pragma: no cover
            if not quiet:  # pragma: no cover
                print("Can't open/read: {e}".format(e=e))  # noqa: T201 - it's fine here; # pragma: no cover

    stash.Finalize()

    cached_res = state._caches.GetFromCache(
        [x.ID for x in rules], stash.FingerPrint)
    if cached_res is not None:
        return (cached_res, {})

    inline_supp_map = {}
    for item in stash.GetItemsFor(classifier=Comment.CLASSIFIER):
        for index, line in enumerate(item.get_items()):
            m = re.match(
                r'^#\s+nooelint:\s+(?P<ids>[A-Za-z0-9\.,_\s-]*)', line, re.MULTILINE)
            if m:
                if item.Origin not in inline_supp_map:  # pragma: no cover
                    inline_supp_map[item.Origin] = {}
                pos = item.InFileLine + index
                inline_supp_map[item.Origin][pos] = [
                    x.strip() for x in re.split(r',|\s+', m.group('ids')) if x]

    state.inline_suppressions = {
        **state.inline_suppressions, **inline_supp_map}

    rules = [copy.deepcopy(x) for x in rules]
    for rule in rules:
        rule.set_state(state)
        rule.set_product_matrix(matrix)
        rule.set_rungroup(group_files)

    _files = list(set(stash.GetRecipes() + stash.GetLoneAppends() +
                  stash.GetConfFiles() + stash.GetBBClasses()))
    issues = []
    for _, f in enumerate(_files):
        _classification = Rule.classify_file(f)
        for r in rules:
            if not r.should_run(_classification):
                continue
            if fix:
                fixedfiles += r.fix(f, stash)
            issues += r.check(f, stash)
    fixedfiles = list(set(fixedfiles))
    fixedfiledict = {}
    for f in fixedfiles:
        items: List[Item] = stash.GetItemsFor(filename=f)
        for file in {x.Origin for x in items}:  # noqa: VNE002
            cnt = ''.join([x.RealRaw for x in sorted(
                items, key=lambda key: key.InFileLine) if x.Origin == file])
            if persistent:
                if not state.nobackup:
                    os.rename(file, file + '.bak')  # pragma: no cover
                with open(file, 'w') as o:
                    o.write(cnt)
                    if not quiet:
                        print('{path}:{lvl}:{msg}'.format(path=os.path.abspath(file),  # noqa: T201 - it's fine here; # pragma: no cover
                              lvl='debug', msg='Applied automatic fixes'))
            else:
                fixedfiledict[file] = cnt  # pragma: no cover

    if any(isinstance(x, FileNotApplicableInlineSuppression) for x in rules):
        known_ids = list(itertools.chain(*[x.get_ids() for x in rules]))
        for _file, _lineobj in inline_supp_map.items():
            for _line, _ids in _lineobj.items():
                for _id in _ids:
                    if not state.get_inline_suppression_seen(_file, _line, _id):
                        if _id not in known_ids:
                            continue
                        obj = FileNotApplicableInlineSuppression(state)
                        issues += obj.finding(_file, _line,
                                              override_msg=obj.Msg.format(id=_id))

    state._caches.SaveToCache([x.ID for x in rules], stash.FingerPrint, issues)

    return (issues, fixedfiledict)


def flatten(list_: Iterable) -> Tuple:
    if not isinstance(list_, list):
        return (list_, {})  # pragma: no cover
    res_a = []
    res_b = {}
    for sublist in list_:
        _a, _b = sublist
        res_a.extend(_a)
        res_b = {**res_b, **_b}
    return (res_a, res_b)


def create_lib_arguments(files: List[str],
                         suppressions: List[str] = None,
                         fix: bool = False,
                         nobackup: bool = False,
                         additional_rules: List[str] = None,
                         rulefile: str = None,
                         jobs: int = None,
                         color: bool = False,
                         quiet: bool = False,
                         noinfo: bool = False,
                         nowarn: bool = False,
                         hide: List[str] = None,
                         relpaths: bool = False,
                         messageformat: str = None,
                         constantmods: List[str] = None,
                         release: str = None,
                         mode: str = 'fast',
                         cached: bool = False,
                         cachedir: str = __default_cache_dir,
                         extra_layer: List[str] = None) -> argparse.Namespace:
    """Create runtime arguments in library mode

    Args:
        files (List[str]): List of files to lint
        suppressions (List[str], optional): List of suppressions. Defaults to None.
        fix (bool, optional): Automatically fix fixable issues. Defaults to False.
        nobackup (bool, optional): No backup for fixed files. Defaults to False.
        additional_rules (List[str], optional): Path(s) to directories with additional rules. Defaults to None.
        rulefile (str, optional): Path to rulefile. Defaults to None.
        jobs (int, optional): Number of jobs. Defaults to None.
        color (bool, optional): Color output. Defaults to False.
        quiet (bool, optional): Quiet mode. Defaults to False.
        hide (List[str], optional): hide messages of specified severity. Defaults to None.
        noinfo (bool, optional): No info messages. Defaults to False. (legacy interface)
        nowarn (bool, optional): No warning messages. Defaults to False. (legacy interface)
        relpaths (bool, optional): Use relative paths in output. Defaults to False.
        messageformat (str, optional): Override message format. Defaults to None.
        constantmods (List[str], optional): Constant mods. Defaults to None.
        release (str, optional): Release to check against. Defaults to None.
        mode (str, optional): Level of testing. Defaults to fast.
        cached (bool, optional): Use caching
        cachedir (str, optional): Path to cache directory,
        extra_layer (List[str], optional): Extra 3rd party layer to load data for

    Returns:
        argparse.Namespace: runtime arguments
    """
    parser = argparse.ArgumentParser()
    parser.register('action', 'tsappend', TypeSafeAppendAction)
    parser.add_argument('--suppress', default=[])
    parser.add_argument('--fix', action='store_true', default=False)
    parser.add_argument('--nobackup', action='store_true', default=False)
    parser.add_argument('--addrules', nargs='+', default=[])
    parser.add_argument('--customrules', nargs='+', default=[])
    parser.add_argument('--rulefile', default=None)
    parser.add_argument('--quiet', action='store_true', default=False)
    parser.add_argument('--jobs', type=int, default=mp.cpu_count())
    parser.add_argument('--color', action='store_true', default=False)
    parser.add_argument('--hide', action='append', default=None,
                        choices=['info', 'warning', 'error'])
    parser.add_argument('--noinfo', action='store_true', default=False)
    parser.add_argument('--nowarn', action='store_true', default=False)
    parser.add_argument('--relpaths', action='store_true', default=False)
    parser.add_argument(
        '--messageformat', default='{path}:{line}:{severity}:{id}:{msg}', type=str)
    parser.add_argument('--constantmods', default=[], nargs='+')
    parser.add_argument(
        '--release', default=Tweaks.DEFAULT_RELEASE, choices=Tweaks.releases())
    parser.add_argument('--mode', default='fast', choices=['fast', 'all'])
    parser.add_argument('--cached', action='store_true', help='Use caches')
    parser.add_argument('--cachedir', default=os.environ.get('OELINT_CACHE_DIR', __default_cache_dir),
                        help=f'Cache directory (default {__default_cache_dir})')
    parser.add_argument('--clear-caches', action='store_true',
                        help='Clear cache directory and exit')
    parser.add_argument('--extra-layer', nargs='*', action='extend',
                        default=['core'], help='Layer names of 3rd party layers to use')
    # Override the defaults with the values from the config file
    parser.set_defaults(**parse_configfile())

    parser.add_argument('files', nargs='*')

    dummy_args = [y for y in [
        *[f'--suppress={x}' for x in (suppressions or ())],
        '--fix' if fix else '',
        '--nobackup' if nobackup else '',
        *[f'--addrules={x}' for x in (additional_rules or ())],
        f'--rulefile={rulefile}' if rulefile else '',
        f'--jobs={jobs}' if jobs else '',
        '--color' if color else '',
        '--quiet' if quiet else '',
        '--noinfo' if noinfo else '',
        '--nowarn' if nowarn else '',
        *[f'--hide={x}' for x in (hide or ())],
        '--relpaths' if relpaths else '',
        f'--messageformat={messageformat}' if messageformat else '',
        *[f'--constantmods={x}' for x in (constantmods or ())],
        f'--release={release}' if release else '',
        f'--mode={mode}',
        '--cached' if cached else '',
        f'--cachedir={cachedir}',
        *[f'--extra-layer={x}' for x in (extra_layer or ())],
        * files,
    ] if y != '']

    return arguments_post(parser.parse_args(dummy_args))


def arguments_post(args: argparse.Namespace) -> argparse.Namespace:  # noqa: C901 - complexity is still okay
    setattr(args, 'state', State())  # noqa: B010

    # Convert boolean symbols
    for _option in [
        'color',
        'exit_zero',
        'fix',
        'nobackup',
        'noinfo',
        'nowarn',
        'print_rulefile',
        'quiet',
        'relpaths',
    ]:
        try:
            setattr(args, _option, bool(getattr(args, _option)))
        except AttributeError:  # pragma: no cover
            pass  # pragma: no cover

    # Convert list symbols
    for _option in [
        'addrules',
        'constantmods',
        'customrules',
        'suppress',
        'extra_layer',
        'hide',
    ]:
        try:
            if not isinstance(getattr(args, _option), list):
                setattr(args, _option, [x.strip() for x in (
                    getattr(args, _option) or '').split('\n') if x])
        except AttributeError:  # pragma: no cover
            pass  # pragma: no cover

    # Apply release specific tweaks
    args = Tweaks.tweak_args(args)

    # adjust potentially relative paths
    args.customrules = [os.path.join(config_basepath, x)
                        for x in args.customrules]

    if args.files == [] and not args.print_rulefile and not args.clear_caches:
        raise argparse.ArgumentTypeError('no input files')

    if args.rulefile:
        args.rulefile = os.path.join(config_basepath, args.rulefile)
        try:
            with open(args.rulefile) as i:
                args.state.rule_file = json.load(i)
        except (FileNotFoundError, json.JSONDecodeError):
            raise argparse.ArgumentTypeError(
                "'rulefile' is not a valid file")

    if args.hide:
        for severity in args.hide:
            args.state.hide[severity] = True

    if args.noinfo:
        args.state.hide['info'] = True

    if args.nowarn:
        args.state.hide['warning'] = True

    args.state._caches = Caches(args)

    for mod in args.constantmods:
        if isinstance(mod, str):
            try:
                with open(os.path.join(config_basepath, mod.lstrip('+-'))) as _in:
                    _cnt = json.load(_in)
                if mod.startswith('+'):
                    CONSTANTS.AddConstants(_cnt)
                    args.state._caches.AddToFingerPrint(f'+{_cnt}')
                elif mod.startswith('-'):
                    CONSTANTS.RemoveConstants(_cnt)
                    args.state._caches.AddToFingerPrint(f'-{_cnt}')
                else:
                    CONSTANTS.OverrideConstants(_cnt)
                    args.state._caches.AddToFingerPrint(str(_cnt))
            except (FileNotFoundError, json.JSONDecodeError):
                raise argparse.ArgumentTypeError(
                    "mod file '{file}' is not a valid file".format(file=mod))
        else:
            CONSTANTS.AddConstants(mod.get('+', {}))
            CONSTANTS.RemoveConstants(mod.get('-', {}))
            args.state._caches.AddToFingerPrint(f'mod{mod.get("+", {})}')
            args.state._caches.AddToFingerPrint(f'mod{mod.get("-", {})}')

    args.state.color = args.color
    args.state.nobackup = args.nobackup
    args.state.rel_path = args.relpaths
    args.state.suppression = args.suppress

    args.state.messageformat = args.messageformat

    if args.fix and args.jobs > 1:
        args.jobs = 1
        print('WARNING: --fix should only be run in single job mode (--jobs=1) - downgrading to 1 job')  # noqa: T201

    return args


def run(args: argparse.Namespace, persistent: bool = True) -> Tuple[List[Tuple[str, int, str]], Dict[str, str]]:
    rules = load_rules(args, add_rules=args.addrules,
                       add_dirs=args.customrules)
    _loaded_ids = []
    _rule_file = args.state.get_rulefile()

    def rule_applicable(rule):
        if isinstance(rule, Rule):
            res = not _rule_file or any(
                x in _rule_file for x in rule.get_ids())  # pragma: no cover
            res &= rule.ID not in args.state.get_suppressions()
            res &= (rule.get_severity() in ['info', 'warning', 'error']) or any(
                rule.get_severity(x) in ['info', 'warning', 'error'] for x in rule.Appendix)
        else:
            res = not _rule_file or rule in _rule_file
            res &= rule not in args.state.get_suppressions()
        return res

    rules = [x for x in rules if rule_applicable(x)]

    for r in rules:
        _loaded_ids += [x for x in r.get_ids() if rule_applicable(x)]
    if not args.quiet:
        print('Loaded rules:\n\t{rules}'.format(  # noqa: T201 - it's here for a reason
            rules='\n\t'.join(sorted(_loaded_ids))))
    issues = []
    fixedfiles = {}
    groups = group_files(args.files, args.mode)
    if not any(groups):
        return ([], {})
    # Starting with python 3.14 the default way of starting the mp Pool
    # will be 'forkserver' - see https://docs.python.org/3.14/library/multiprocessing.html#contexts-and-start-methods
    # But in our setup we will need to share e.g. the CONSTANTS object into
    # the checks running in the Pool
    # Hence, we enforce fork
    ctx = mp.get_context('fork')
    with ctx.Pool(processes=min(args.jobs, len(groups))) as pool:
        try:
            issues, fixedfiles = flatten(pool.map(partial(group_run, quiet=args.quiet, fix=args.fix,
                                                          rules=rules, state=args.state, persistent=persistent), groups))
        finally:
            pool.close()
            pool.join()

    # deduplicate matrix
    def deduplicate(input_list):
        seen = []
        result = set()
        for item in input_list:
            needle = (item[0], item[2])
            if needle not in seen:
                seen.append(needle)
                matrix_str = f' [{",".join(sorted(item[1]))}]' if item[1] else ''
                result.add((item[0], f'{item[2]}{matrix_str}'))
        return result

    issues = sorted(deduplicate(issues), key=lambda x: x[0])

    return (issues, fixedfiles)



import fnmatch
import importlib
import inspect
import os
import pkgutil
import sys
from enum import Enum
from typing import FrozenSet, Iterable, List, Tuple

from colorama import Style
from oelint_parser.cls_stash import Stash

from oelint_adv.state import State
from oelint_adv.version import __version__


class Classification(Enum):
    RECIPE = 0
    BBAPPEND = 1
    BBCLASS = 2
    MACHINECONF = 3
    DISTROCONF = 4
    LAYERCONF = 5

    @staticmethod
    def tostr(val) -> str:  # noqa: VNE002
        if val == Classification.RECIPE:
            return 'recipe'
        if val == Classification.BBAPPEND:
            return 'bbappend'
        if val == Classification.BBCLASS:
            return 'bbclass'
        if val == Classification.MACHINECONF:
            return 'machine configuration'
        if val == Classification.DISTROCONF:
            return 'distro configuration'
        return 'layer configuration'


class Rule:
    def __init__(self,
                 id: str = '',  # noqa: A002, VNE003
                 severity: str = '',
                 message: str = '',
                 onappend: bool = True,
                 onlyappend: bool = False,
                 appendix: List[str] = (),
                 valid_till_release: str = '',
                 valid_from_release: str = '',
                 run_on: List[Classification] = None,
                 ) -> None:  # noqa: A002, VNE003
        """constructor

        Keyword Arguments:
            id {str} -- ID of the rule (default: {''})
            severity {str} -- severity of the rule (default: {''})
            message {str} -- Rule message (default: {''})
            onappend {bool} -- (deprecated: use run_on) true if rule should be run on bbappends (default: {True})
            onlyappend {bool} -- (deprecated: use run_on) true if rule applies to bbappends only (default: {False})
            appendix {List[str]} -- possible appendix to id
            valid_till_release {str} -- rule only valid till (excluding) this release (default: '' = all)
            valid_from_release {str} -- rule only valid from (including) this release (default: '' = all)
            run_on {List[Classification]} -- run rule only on certain file groups (default: None = all)
        """
        self.ID = id
        self.Severity = severity
        self.Msg = message
        self.Appendix = appendix
        self._run_on: List[Classification] = run_on or [
            Classification.BBAPPEND,
            Classification.BBCLASS,
            Classification.DISTROCONF,
            Classification.LAYERCONF,
            Classification.MACHINECONF,
            Classification.RECIPE,
        ]
        self.OnlyAppend = onlyappend
        self.OnAppend = onappend
        if self.OnlyAppend:
            self._run_on = [Classification.BBAPPEND]  # pragma: no cover
        if not self.OnAppend and Classification.BBAPPEND in self._run_on:
            self._run_on.remove(Classification.BBAPPEND)  # pragma: no cover
        self._valid_till_release = valid_till_release
        self._valid_from_release = valid_from_release
        self._state: State = None
        self.__matrix: FrozenSet[str] = []
        self.__rungroup: List[str] = []

    @staticmethod
    def classify_file(fn) -> List[Classification]:
        if fnmatch.fnmatch(fn, '*.bbclass'):
            return [Classification.BBCLASS]
        if fnmatch.fnmatch(fn, '*.bbappend'):
            return [Classification.BBAPPEND]
        if fnmatch.fnmatch(fn, '*.bb'):
            return [Classification.RECIPE]
        if fnmatch.fnmatch(fn, '*/machine/*.conf'):
            return [Classification.MACHINECONF]
        if fnmatch.fnmatch(fn, '*/distro/*.conf'):
            return [Classification.DISTROCONF]
        if fnmatch.fnmatch(fn, '*/layer.conf'):
            return [Classification.LAYERCONF]
        return []  # pragma: no cover

    def set_product_matrix(self, in_: FrozenSet[str]) -> None:
        """Set the product matrix

        Args:
            in_ (frozenset[str]): Product matrix of ids
        """
        self.__matrix = in_

    def set_state(self, state_: State) -> None:
        """Set the state object

        Args:
            state_ (State): Current state object
        """
        self._state = state_

    def set_rungroup(self, files: List[str]) -> None:
        """Set the rungroup aka files that being checked together

        Args:
            files (List[str]): Files being checked as part of the group
        """
        self.__rungroup = files

    def check_release_range(self, release_range: List[str]) -> bool:
        """Check if rule is applicable with currently configured release(s)

        Args:
            release_range (List[str]): Range of supported releases (as passed by tweaks)

        Returns:
            bool: Rule is applicable with release range
        """
        if self._valid_from_release and self._valid_from_release not in release_range:
            return False
        if self._valid_till_release and self._valid_till_release in release_range:
            return False
        return True

    def should_run(self, groupclass: List[Classification]) -> bool:
        """Wrapper for check, checking if the rule classification matches the group

        Args:
            groupclass (List[Classification]): Classifications of the group

        Returns:
            bool: rule should be run
        """
        return any(set(groupclass).intersection(self._run_on))

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        """Stub for running check - is overridden by each rule

        Arguments:
            _file {str} -- File to be parsed
            stash {oelint_adv.cls_stash.Stash} -- Parsed stash

        Returns:
            list -- List of findings
        """
        return []  # pragma: no cover

    def fix(self, _file: str, stash: Stash) -> List[str]:
        """Stub for fix function - can be overridden by each rule

        Arguments:
            _file {str} -- File to be parsed
            stash {oelint_adv.cls_stash.Stash} -- Parsed stash

        Returns:
            list -- list of changed files
        """
        return []

    def finding(self,
                _file: str,
                _line: int,
                override_msg: str = None,
                appendix: str = None,
                blockoffset: int = 0,
                severity_override: str = '') -> Tuple[Tuple[str, int], List[str], str]:
        """Called by rule to indicate a finding

        Arguments:
            _file {str} -- Full path to file or origin
            _line {int} -- Line number in file

        Keyword Arguments:
            override_msg {str} -- Optional string which overrides the set standard message (default: {None})
            appendix {str} -- Optional appendix to rule ID (default: {None})
            blockoffset {int} -- line number to look for inline suppressions instead of _line (default: 0 == use _line)
            severity_override {str} -- override the base severity (empty == use base)

        Returns:
            Tuple[Tuple[str, int], List[str], str] -- Path, line, matrix, Human readable finding (possibly with color codes)
        """
        if override_msg is None:
            override_msg = self.Msg
        _suppression_offset = blockoffset or _line
        _id = [self.ID]
        _display_id = self.ID
        if appendix:
            _id.append(self.ID + '.' + appendix)
            _display_id += '.' + appendix
        # filter out suppressions
        if any(x in self._state.get_suppressions() for x in _id):
            return []
        _severity = self.get_severity(appendix) or severity_override

        if self._state.get_hide(_severity):
            return []

        if _line <= 0:
            # Fix those issues, that don't come with a line
            _line = 1

        # filter out inline suppressions
        if any(x for x in _id if x in self._state.get_inlinesuppressions().get(_file, {}).get(max(1, _suppression_offset - 1), [])):
            self._state.set_inline_suppression_seen(
                _file, _suppression_offset - 1, _id)
            return []

        _path = os.path.abspath(_file)
        if self._state.get_relpaths():
            _path = os.path.relpath(_path, os.getcwd())

        _style = ''
        _color = ''
        if self._state.get_colorize():
            _color = self._state.get_color_by_severity(_severity)
            _style = Style.RESET_ALL

        wikiurl = f'https://github.com/priv-kweihmann/oelint-adv/blob/{__version__}/docs/wiki/{self.ID}.md'

        _msg = self._state.get_messageformat().format(path=_path, line=_line, severity=_severity,
                                                      id=_display_id, msg=override_msg,
                                                      wikiurl=wikiurl,
                                                      rungroup=','.join(self.__rungroup))
        return [((_path, _line), self.__matrix, f'{_color}{_msg}{_style}')]

    def __repr__(self) -> str:
        return '{id}'.format(id=self.ID)  # pragma: no cover

    def get_severity(self, appendix: str = None) -> str:
        """Get the configured severity for this rule, if it is enabled.

        Keyword Arguments:
            appendix {str} -- Potential subrule name (default: {None})

        Returns:
            str -- Severity for this rule if it is enabled, {None} if disabled.
        """
        _rule_file = self._state.get_rulefile()
        if not _rule_file:
            return self.Severity
        _subid = None if appendix is None else f'{self.ID}.{appendix}'
        if _subid and _subid in _rule_file:
            _severity = _rule_file[_subid]
        elif self.ID in _rule_file:
            _severity = _rule_file[self.ID]
        else:
            # rule not in rulefile
            return None
        return _severity if _severity != '' else self.Severity

    def get_ids(self) -> List[str]:
        """Returns all possible IDs of the rule

        Returns:
            list -- possible IDS of the rule
        """
        return [self.ID] + ['{id}.{app}'.format(id=self.ID, app=x) for x in self.Appendix]

    def get_rulefile_entries(self) -> dict:
        """Returns a dictionary of entries which would represent the currently
        enabled ruleset (for this rule) in a rulefile.

        Returns:
            dict -- list of rulefile entries
        """
        return {
            **({} if (self.get_severity() is None or self.ID in self._state.get_suppressions()) else {self.ID: self.get_severity()}),
            **{f'{self.ID}.{x}': self.get_severity(x) for x in self.Appendix if (self.get_severity(x) is not None and self.ID not in self._state.get_suppressions())},
        }

    def format_message(self, *args, **kwargs) -> str:
        """Format message

        Returns:
            str -- formatted message
        """
        return self.Msg.format(*args, **kwargs)

    @staticmethod
    def is_lone_append(stash: Stash, file: str) -> bool:  # noqa: VNE002
        """Check if the file is a bbappend file without any matching
           bb file

        Args:
            stash {oelint_parser.cls_stash.Stash} -- Parsed stash
            file {str} -- Path to file

        Returns:
            bool: True if bbappend has no matches in stash
        """
        if not file.endswith('.bbappend'):
            return False
        return file in stash.GetLoneAppends()


def load_rules(args, add_rules: Iterable[str] = (), add_dirs: Iterable[str] = ()) -> List['Rule']:
    """Load rules from set directories

    Keyword Arguments:
        add_rules {tuple} -- Additional builtin rulesets to be loaded (default: {[]})
        add_dirs {tuple} -- Additional directories to parse for rules (default: {[]})

    Returns:
        list -- Class instances of loaded rules
    """
    res = []
    _rule_file = args.state.get_rulefile()
    _path_list = {
        'base': {'path': 'rule_base', 'builtin': True},
    }
    for ar in add_rules:
        _path_list[ar] = {'path': 'rule_{a}'.format(a=ar), 'builtin': True}
    for ar in add_dirs:
        _path_list['additional_{a}'.format(a=os.path.basename(ar))] = {
            'path': ar, 'builtin': False}  # pragma: no cover
    for _, v in _path_list.items():
        if v['builtin']:
            _searchpath = os.path.join(os.path.dirname(
                os.path.abspath(__file__)), v['path'])
        else:
            _searchpath = os.path.join(v['path'])  # pragma: no cover
            sys.path.append(os.path.dirname(v['path']))  # pragma: no cover
        packages = pkgutil.walk_packages(path=[_searchpath])
        for _, name, _ in packages:
            if v['builtin']:
                name = __name__.split('.')[0] + '.' + v['path'] + '.' + name
            else:
                name = os.path.basename(v['path']) + \
                    '.' + name  # pragma: no cover
            try:
                mod = importlib.import_module(name)
                for m in inspect.getmembers(mod, inspect.isclass):
                    try:
                        if issubclass(m[1], Rule):
                            inst = m[1]()
                            if not inst.check_release_range(args._release_range):
                                continue
                            inst._state = args.state
                            _potential_ids = [
                                inst.ID] + ['{a}.{b}'.format(a=inst.ID, b=x) for x in inst.Appendix]
                            if any(_potential_ids):
                                if _rule_file and not any(x in _rule_file for x in _potential_ids):
                                    continue  # pragma: no cover
                                res.append(inst)
                    except Exception:  # pragma: no cover, # noqa: S110
                        pass  # pragma: no cover
            except Exception as e:  # pragma: no cover
                if not args.quiet:  # pragma: no cover
                    print(  # noqa: T201 this print is fine here
                        "Can't load rule {rule} -> {exp}".format(rule=name, exp=e))  # pragma: no cover
    return res



import os
from typing import List, Tuple

import regex  # noqa: I900
from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash
from oelint_parser.rpl_regex import RegexRpl

from oelint_adv.cls_rule import Rule, Classification


class FilePatchUpstreamStatusOccurance(Rule):
    def __init__(self) -> None:
        self._valid_class = {
            'Pending': r'^Upstream-Status:\s*Pending',
            'Submitted': r'^Upstream-Status:\s*Submitted(\s+\[.*\])*',
            'Accepted': r'^Upstream-Status:\s*Accepted',
            'Denied': r'^Upstream-Status:\s*Denied',
            'Backport': r'^Upstream-Status:\s*Backport',
            'Inappropriate': r'^Upstream-Status:\s*Inappropriate(\s+\[.*\])*',
            'Inactive-Upstream': r'^Upstream-Status:\s*Inactive-Upstream(\s+\[.*\])*',
        }
        super().__init__(id='oelint.file.upstreamstatus_occurance',
                         severity='inactive',
                         run_on=[Classification.BBAPPEND, Classification.RECIPE],
                         message="Found {ID} set as Upstream-Status in patch '{FILE}'",
                         appendix=list(self._valid_class.keys()))

    def _get_recipe(self, items, path):
        # Find matching SRC_URI assignment
        return [x for x in items if x.VarValue.find(os.path.basename(path)) != -1]

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        patches = stash.GetFiles(_file, '*.patch')
        _items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                   attribute=Variable.ATTR_VAR, attributeValue='SRC_URI')

        for i in patches:
            with open(i) as _input:
                _recipe_match = self._get_recipe(_items, i)
                if _recipe_match:
                    _recipe_match = _recipe_match[0]
                else:
                    continue  # pragma: no cover
                try:
                    cnt = _input.read()
                    for k, v in self._valid_class.items():
                        for _ in RegexRpl.finditer(v, cnt, flags=regex.regex.MULTILINE):
                            res += self.finding(_recipe_match.Origin,
                                                _recipe_match.InFileLine,
                                                override_msg=self.Msg.format(ID=k, FILE=os.path.basename(i)),
                                                appendix=k)
                except UnicodeDecodeError:  # pragma: no cover
                    pass  # pragma: no cover
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Include
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class FileIncludeVsRequire(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.file.requireinclude',
                         severity='warning',
                         message="Use 'require {FILE}' instead of 'include {FILE}'")

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        for item in stash.GetItemsFor(filename=_file,
                                      classifier=Include.CLASSIFIER):
            if item.Statement == 'include':
                res += self.finding(item.Origin, item.InFileLine,
                                    self.Msg.replace('{FILE}', item.IncName))
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash
from oelint_parser.parser import INLINE_BLOCK

from oelint_adv.cls_rule import Rule


class VarSRCURIWildcard(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.var.srcuriwildcard',
                         severity='error',
                         message="'SRC_URI' should not contain any wildcards")

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        _items: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                   attribute=Variable.ATTR_VAR, attributeValue='SRC_URI')
        for item in _items:
            for f in [x.strip("'") for x in item.get_items() if x and INLINE_BLOCK not in x]:
                components = stash.GetScrComponents(f)
                if components['scheme'] == 'file':
                    if any(x for x in ['*'] if x in components['src']):
                        res += self.finding(item.Origin, item.InFileLine)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class VarQuoted(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.valuequoted',
                         severity='error',
                         message='Variable value should be quoted')

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER)
        for i in items:
            # Don't use VarValueStripped here as we explicitly want the quotes
            # at the beginning and the end of the value
            val_ = i.VarValue.strip()
            if ((not val_.startswith('"') or not val_.endswith('"'))
                    and (not val_.startswith("'") or not val_.endswith("'"))):  # noqa: W503
                res += self.finding(i.Origin, i.InFileLine)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import FunctionExports, Inherit, Item, Variable
from oelint_parser.cls_stash import Stash
from oelint_parser.constants import CONSTANTS

from oelint_adv.cls_rule import Classification, Rule


class VarOutOfContext(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.outofcontext',
                         severity='warning',
                         message='{var} should be only set in {context}')

    def _check(self,
               results: List[Tuple[str, int, str]],
               items: List[Item],
               valid_context: List[Classification],
               var_desc: str,
               context_desc: str = '') -> None:
        for item in items:
            classification = Rule.classify_file(item.Origin) or [
                Classification.RECIPE]
            if not any(set(classification).intersection(valid_context)):
                context = context_desc or ",".join(
                    f'{Classification.tostr(x)}(s)' for x in valid_context)
                results += self.finding(item.Origin, item.InFileLine,
                                        override_msg=self.Msg.format(var=var_desc, context=context))

    def _is_image(self, _file: str, stash: Stash):
        res = False
        _inherits = stash.GetItemsFor(
            filename=_file, classifier=Inherit.CLASSIFIER)
        for item in _inherits:
            res |= any(True for x in item.get_items()
                       if x in CONSTANTS.ImagesClasses)
        return res

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []

        for var in CONSTANTS.GetByPath('oelint-contextvars/conf-only'):  # noqa: VNE002
            self._check(res, stash.GetItemsFor(filename=_file,
                                               classifier=Variable.CLASSIFIER,
                                               attribute=Variable.ATTR_VAR,
                                               attributeValue=var),
                        [Classification.LAYERCONF, Classification.MACHINECONF,
                            Classification.DISTROCONF],
                        var_desc=var)
        for var in CONSTANTS.GetByPath('oelint-contextvars/bbappend-only'):  # noqa: VNE002
            self._check(res, stash.GetItemsFor(filename=_file,
                                               classifier=Variable.CLASSIFIER,
                                               attribute=Variable.ATTR_VAR,
                                               attributeValue=var),
                        [Classification.BBAPPEND],
                        var_desc=var)

        for var in CONSTANTS.GetByPath('oelint-contextvars/bbclass-only'):  # noqa: VNE002
            self._check(res, stash.GetItemsFor(filename=_file,  # pragma: no cover
                                               classifier=Variable.CLASSIFIER,
                                               attribute=Variable.ATTR_VAR,
                                               attributeValue=var),
                        [Classification.BBCLASS],
                        var_desc=var)

        for var in CONSTANTS.GetByPath('oelint-contextvars/recipe-only'):  # noqa: VNE002
            self._check(res, stash.GetItemsFor(filename=_file,
                                               classifier=Variable.CLASSIFIER,
                                               attribute=Variable.ATTR_VAR,
                                               attributeValue=var),
                        [Classification.BBCLASS, Classification.BBAPPEND,
                            Classification.RECIPE],
                        var_desc=var)
        for var in CONSTANTS.GetByPath('oelint-contextvars/image-only'):  # noqa: VNE002
            if self._is_image(_file, stash):
                continue
            self._check(res, stash.GetItemsFor(filename=_file,
                                               classifier=Variable.CLASSIFIER,
                                               attribute=Variable.ATTR_VAR,
                                               attributeValue=var),
                        [Classification.MACHINECONF, Classification.DISTROCONF],
                        var_desc=var,
                        context_desc='file(s) that define image(s)')

        self._check(res, stash.GetItemsFor(filename=_file,
                                           classifier=FunctionExports.CLASSIFIER),
                    [Classification.BBCLASS],
                    var_desc='EXPORT_FUNCTIONS')

        self._check(res, stash.GetItemsFor(filename=_file,
                                           classifier=Inherit.CLASSIFIER,
                                           attribute=Inherit.ATTR_CLASS,
                                           attributeValue=['native', 'nativesdk', 'cross']),
                    [Classification.BBCLASS, Classification.BBAPPEND,
                        Classification.RECIPE],
                    var_desc='inherit native, nativesdk, cross')

        return res



import fnmatch
from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule, Classification


class VarsDistroConf(Rule):
    def __init__(self) -> None:
        self.needles = [
            'MACHINE',
            'MACHINE_ARCH',
            'MACHINE_ESSENTIAL_EXTRA_RDEPENDS',
            'MACHINE_ESSENTIAL_EXTRA_RRECOMMENDS',
            'MACHINE_EXTRA_RRECOMMENDS',
            'MACHINE_FEATURES',
            'MACHINE_FEATURES_BACKFILL',
            'MACHINE_FEATURES_BACKFILL_CONSIDERED',
            'IMAGE_INSTALL',
            'MACHINEOVERRIDES',
        ]

        super().__init__(id='oelint.vars.distroconf',
                         severity='warning',
                         run_on=[Classification.DISTROCONF],
                         message='{var} should not be set as part of a distro configuration',
                         appendix=self.needles)

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                  attribute=Variable.ATTR_VAR)

        for i in items:
            if not fnmatch.fnmatch(i.Origin, '*/distro/*.conf'):
                continue   # pragma: no cover
            if i.VarName in self.needles:
                res += self.finding(_file, i.InFileLine,
                                    self.Msg.format(var=i.VarName), appendix=i.VarName)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class VarListAppend(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.listappend',
                         severity='error',
                         message='<FOO>')

    def __getMatches(self, _file: str, stash: Stash) -> Tuple[List[Variable], List[Variable]]:
        res_app = []
        res_pre = []
        items: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                  attribute=Variable.ATTR_VAR,
                                                  attributeValue=['PACKAGES', 'SRC_URI', 'FILES', 'RDEPENDS', 'DEPENDS'])

        for i in items:
            ops = i.AppendOperation()
            if not i.VarValue.startswith('" ') and any(x in ops for x in ['append', ' .= ']):
                res_app.append(i)
            if not i.VarValue.endswith(' "') and any(x in ops for x in ['prepend', ' =. ']):
                res_pre.append(i)
        return (res_app, res_pre)

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        res_app, res_pre = self.__getMatches(_file, stash)
        for i in res_app:
            res += self.finding(i.Origin, i.InFileLine,
                                override_msg='Append to list should start with a blank')
        for i in res_pre:
            res += self.finding(i.Origin, i.InFileLine,
                                override_msg='Prepend to list should end with a blank')
        return res

    def fix(self, _file: str, stash: Stash) -> List[str]:
        res = []
        res_app, res_pre = self.__getMatches(_file, stash)
        for i in res_app:
            vv_rpl = i.VarValue.replace('"', '" ', 1).replace("'", "' ", 1)
            i.RealRaw = i.RealRaw.replace(i.VarValue, vv_rpl)
            i.Raw = i.Raw.replace(i.VarValue, vv_rpl)
            i.VarValue = vv_rpl
            res.append(_file)
        for i in res_pre:
            vv_rpl = i.VarValue[::-1].replace('"', '" ', 1).replace("'", "' ", 1)[::-1]
            i.RealRaw = i.RealRaw.replace(i.VarValue, vv_rpl)
            i.Raw = i.Raw.replace(i.VarValue, vv_rpl)
            i.VarValue = vv_rpl
            res.append(_file)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Function
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class TaskInstallNoMkdir(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.task.nomkdir',
                         severity='error',
                         message="'mkdir' shall not be used in do_install. Use 'install'")

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Function] = stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER)
        for item in items:
            if item.FuncName.startswith('do_install'):
                for lineindex, line in enumerate(item.get_items()):
                    if line.startswith('mkdir ') or line.find(' mkdir ') != -1:
                        res += self.finding(item.Origin, item.InFileLine + lineindex, blockoffset=item.InFileLine)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash
from oelint_parser.parser import INLINE_BLOCK

from oelint_adv.cls_rule import Rule


class VarSRCUriSRCREVTag(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.srcurisrcrevtag',
                         severity='error',
                         message="'tag' in SRC_URI and a SRCREV for the same component doesn't compute")

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                  attribute=Variable.ATTR_VAR, attributeValue='SRC_URI')
        srcrevs: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                    attribute=Variable.ATTR_VAR, attributeValue='SRCREV')
        for item in items:
            lines = [y.strip('"') for y in item.get_items() if y and INLINE_BLOCK not in y]
            for x in lines:
                _url = stash.GetScrComponents(x)
                if _url['scheme'] in ['git'] and 'tag' in _url['options']:
                    if 'name' in _url['options']:
                        _srcrevs = [x for x in srcrevs if _url['options']['name'] in x.SubItems]
                    else:
                        _srcrevs = [x for x in srcrevs if not x.SubItems]
                    if any(_srcrevs):
                        res += self.finding(item.Origin, item.InFileLine)
        return res
